# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230414

import json
import subprocess
import time

from neetbox.daemon.client._action_agent import _NeetActionManager as NeetActionManager
from neetbox.daemon.client._client import connection
from neetbox.daemon.client._update_thread import connect_daemon
from neetbox.daemon.server.daemonable_process import DaemonableProcess
from neetbox.logging.formatting import LogStyle
from neetbox.logging.logger import Logger

logger = Logger(style=LogStyle(with_datetime=False, skip_writers=["ws"]))


def __attach_daemon(daemon_config):
    if not daemon_config["allowIpython"]:
        try:
            eval("__IPYTHON__")
        except NameError:
            pass
        else:
            logger.log(
                "NEETBOX DAEMON won't start when debugging in ipython console. If you want to allow daemon run in "
                "ipython, try to set 'allowIpython' to True."
            )
            return False  # ignore if debugging in ipython
    _is_daemon_server_online = connect_daemon()  # try to connect daemon
    logger.debug("daemon connection status: " + str(_is_daemon_server_online))
    if not _is_daemon_server_online:  # if no daemon online
        # check if possible to launch
        if daemon_config["host"] not in ["localhost", "127.0.0.1", "0.0.0.0"]:
            # daemon not running on localhost
            logger.err(
                f"No daemon running at {daemon_config['host']}:{daemon_config['port']}, daemon will not be attached. Continue anyway."
            )
            return False

        logger.log(
            f"No daemon running at {daemon_config['host']}:{daemon_config['port']}, trying to create daemon..."
        )

        popen = DaemonableProcess(
            target="neetbox.daemon.server._daemon_launcher",
            args=["--config", json.dumps(daemon_config)],
            mode=daemon_config["mode"],
            redirect_stdout=subprocess.DEVNULL if daemon_config["mute"] else None,
            env_append={"NEETBOX_DAEMON_PROCESS": "1"},
        ).start()

        time.sleep(1)

        _retry_timeout = 10
        _time_begin = time.perf_counter()

        logger.log("Created daemon process, trying to connect to daemon...")
        while time.perf_counter() - _time_begin < 10:  # try connect daemon
            if connect_daemon():
                return True
            else:
                exit_code = popen.poll()
                if exit_code is not None:
                    logger.err(f"Daemon process exited unexpectedly with exit code {exit_code}.")
                    return False

                time.sleep(0.5)

        logger.err(
            f"Failed to connect to daemon after {_retry_timeout}s, daemon connector won't start."
        )
        return False


def _try_attach_daemon():
    from neetbox.config import get_module_level_config

    _cfg = get_module_level_config()
    if _cfg["enable"]:
        __attach_daemon(_cfg)


action = NeetActionManager.register
ws_subscribe = connection.ws_subscribe
__all__ = ["action", "ws_subscribe", "NeetActionManager", "_try_attach_daemon"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230417

from neetbox.pipeline._signal_and_slot import listen, watch

__all__ = ["watch", "listen"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230414


from neetbox.config import get_module_level_config
from neetbox.daemon._protocol import *
from neetbox.daemon.client._client import connection


def get_base_addr():
    __cfg = get_module_level_config()
    daemon_address = f"{__cfg['host']}:{__cfg['port']}"
    return f"http://{daemon_address}/"


def _fetch(url):
    r = connection.http.get(url)
    _data = r.json()
    return _data


def get_list():
    api_addr = f"{get_base_addr()}{FRONTEND_API_ROOT}/list"
    return _fetch(api_addr)


def get_status_of(name):
    api_addr = f"{get_base_addr()}{FRONTEND_API_ROOT}/status/{name}"
    return _fetch(api_addr)



from dataclasses import dataclass
from typing import Any

FRONTEND_API_ROOT = "/web"
CLIENT_API_ROOT = "/cli"


EVENT_TYPE_NAME_KEY = "event-type"
EVENT_ID_NAME_KEY = "event-id"
NAME_NAME_KEY = "name"
PAYLOAD_NAME_KEY = "payload"


@dataclass
class WsMsg:
    name: str
    event_type: str
    payload: Any
    event_id: int = -1

    def json(self):
        return {
            NAME_NAME_KEY: self.name,
            EVENT_TYPE_NAME_KEY: self.event_type,
            EVENT_ID_NAME_KEY: self.event_id,
            PAYLOAD_NAME_KEY: self.payload,
        }



import functools
import inspect
from ast import literal_eval
from threading import Thread
from typing import Callable, Optional

from neetbox.core import Registry
from neetbox.daemon._protocol import *
from neetbox.daemon.client._client import connection
from neetbox.logging import logger
from neetbox.pipeline import watch
from neetbox.pipeline._signal_and_slot import SYSTEM_CHANNEL
from neetbox.utils.mvc import Singleton


class PackedAction(Callable):
    def __init__(
        self,
        function: Callable,
        name: str = None,
        description: str = None,
        blocking: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.function = function
        self.name = name if name else function.__name__
        self.description = description
        self.argspec = inspect.getfullargspec(self.function)
        self.blocking = blocking

    def get_props_dict(self):
        # _arg_dict = {
        #     _arg_name: self.argspec.annotations.get(_arg_name, None)
        #     for _arg_name in self.argspec.args
        # }
        _arg_anno_dict = self.function.__annotations__
        _args = self.argspec.args
        _arg_dict = {}
        for _arg_name in _args:
            _arg_type = _arg_anno_dict.get(_arg_name, any)
            _arg_dict[_arg_name] = _arg_type if isinstance(_arg_type, str) else _arg_type.__name__
        return {
            "description": self.description,
            "args": _arg_dict,
            "blocking": self.blocking,
        }

    def __call__(self, **argv):
        self.function(argv)  # ignore blocking

    def eval_call(self, params: dict):
        eval_params = dict((k, literal_eval(v)) for k, v in params.items())
        return self.function(**eval_params)


class _NeetActionManager(metaclass=Singleton):
    __ACTION_POOL: Registry = Registry("__NEET_ACTIONS")

    @staticmethod
    def get_action_names():
        return {
            n: _NeetActionManager.__ACTION_POOL[n].argspec
            for n in _NeetActionManager.__ACTION_POOL.keys()
        }

    @staticmethod
    def get_action_dict():
        return {
            name: _NeetActionManager.__ACTION_POOL[name].get_props_dict()
            for name in _NeetActionManager.__ACTION_POOL.keys()
        }

    @staticmethod
    def eval_call(name: str, params: dict, callback: Optional[Callable] = None):
        if name not in _NeetActionManager.__ACTION_POOL:
            logger.err(f"Could not find action with name {name}, action stopped.")
            return False
        target_action = _NeetActionManager.__ACTION_POOL[name]
        logger.log(
            f"Agent runs function '{target_action.name}', blocking = {target_action.blocking}"
        )

        def run_and_callback():
            try:
                returned_data = target_action.eval_call(params)
            except Exception as e:
                returned_data = e
            if callback:
                callback(returned_data)

        if not target_action.blocking:  # non-blocking run in thread
            Thread(
                target=run_and_callback,
                daemon=True,
            ).start()
            return
        else:  # blocking run
            run_and_callback()
            return

    @watch(name="__action", initiative=True, _channel=SYSTEM_CHANNEL)
    def _update_action_dict():
        # for status updater
        return _NeetActionManager.get_action_dict()

    def register(name: Optional[str] = None, description: str = None, blocking: bool = False):
        return functools.partial(
            _NeetActionManager._register, name=name, description=description, blocking=blocking
        )

    def _register(
        function: Callable, name: str = None, description: str = None, blocking: bool = False
    ):
        if (
            description is None and function.__doc__ is not None
        ):  # parse function doc as description
            description = function.__doc__
            if description:
                _description_lines = []
                for _line in description.split("\n"):
                    if len(_line):  # remove empty lines
                        _description_lines.append(_line)
                # find shortest lstrip
                min_lstrip = 99999
                for _line in _description_lines[1:]:  # skip first line
                    min_lstrip = min(len(_line) - len(_line.lstrip()), min_lstrip)
                _parsed_description = _description_lines[0] + "\n"
                for _line in _description_lines[1:]:
                    _parsed_description += _line[min_lstrip:] + "\n"
                description = _parsed_description

        packed = PackedAction(
            function=function, name=name, description=description, blocking=blocking
        )
        _NeetActionManager.__ACTION_POOL._register(what=packed, name=packed.name, overwrite=True)
        _NeetActionManager._update_action_dict()  # update for sync
        return function


@connection.ws_subscribe(event_type_name="action")
def __listen_to_actions(msg):
    _payload = msg[PAYLOAD_NAME_KEY]
    _event_id = msg[EVENT_ID_NAME_KEY]
    _action_name = _payload["name"]
    _action_args = _payload["args"]
    _NeetActionManager.eval_call(
        name=_action_name,
        params=_action_args,
        callback=lambda x: connection.ws_send(
            event_type="action",
            payload={"name": _action_name, ("error" if isinstance(x, Exception) else "result"): x},
            event_id=_event_id,
        ),
    )


# example
if __name__ == "__main__":
    import time

    @_NeetActionManager.register(name="some")
    def some(a, b):
        time.sleep(1)
        return f"a = {a}, b = {b}"

    print("registered actions:")
    action_dict = _NeetActionManager.get_action_dict()
    print(action_dict)

    def callback_fun(text):
        print(f"callback_fun print: {text}")

    _NeetActionManager.eval_call(name="some", params={"a": "3", "b": "4"}, callback=callback_fun)
    print("you should see this line first before callback_fun print")
    time.sleep(4)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230417

import importlib
import pkgutil

from neetbox.config import get_module_level_config
from neetbox.core import Registry
from neetbox.integrations.engine import Engine as engine
from neetbox.integrations.engine import get_installed_engines, get_supported_engines
from neetbox.utils.framing import get_frame_module_traceback

__all__ = [
    "engine",
    "get_supported_engines",
    "get_installed_engines",
]

_QUERY_AFTER_LOAD_WORKSPACE = Registry("__INTEGRATION_LOADER_DICT")


def _iter_import_sub_modules():
    _THIS_MODULE = get_frame_module_traceback(1)
    for sub_module_info in pkgutil.iter_modules(_THIS_MODULE.__path__):
        importlib.import_module(f"{_THIS_MODULE.__name__}.{sub_module_info.name}")


def _post_init_workspace():
    _cfg = get_module_level_config()
    if not _cfg["autoload"]:
        return  # do not load if auto load is disabled
    _iter_import_sub_modules()
    for name, fun in _QUERY_AFTER_LOAD_WORKSPACE.items():
        fun()


call_on_workspace_load = _QUERY_AFTER_LOAD_WORKSPACE.register
__all__ = ["call_on_workspace_load"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230413

DEFAULT_WORKSPACE_CONFIG = {
    "name": None,
    "version": None,
    "logging": {"level": "INFO", "logdir": None},
    "pipeline": {
        "updateInterval": 0.5,
    },
    "integrations": {
        "autoload": True,
        "environment": {"hardware": {"monit": True, "interval": 0.5}, "platform": {"monit": True}},
    },
    "daemon": {
        "enable": True,
        "host": "127.0.0.1",
        "port": 20202,
        "displayName": None,
        "allowIpython": False,
        "mute": True,
        "mode": "detached",
        "uploadInterval": 0.5,
    },
}
WORKSPACE_CONFIG: dict = DEFAULT_WORKSPACE_CONFIG.copy()


def update_workspace_config_with(cfg: dict):
    def _update_dict_recursively(self: dict, the_other: dict):
        for _k, _v in the_other.items():
            if type(_v) is dict:  # currently resolving a dict child
                if _k in self:  # dfs merge
                    _update_dict_recursively(self=self[_k], the_other=the_other[_k])
                else:
                    self[_k] = the_other[_k]
            else:  # not a dict, overwriting
                self[_k] = the_other[_k]

    global WORKSPACE_CONFIG
    _update_dict_recursively(WORKSPACE_CONFIG, cfg)


def get_current():
    global WORKSPACE_CONFIG
    return WORKSPACE_CONFIG



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230414

import os

import werkzeug
from rich.console import Console

console = Console()
import logging
import sys
import time
from dataclasses import dataclass
from threading import Thread
from typing import Any, Dict, Tuple

werkzeug_log = logging.getLogger("werkzeug")
werkzeug_log.setLevel(logging.ERROR)  # disable flask http call logs

import setproctitle
from flask import abort, json, request, send_from_directory
from websocket_server import WebsocketServer

from neetbox.daemon._protocol import *

__PROC_NAME = "NEETBOX SERVER"
setproctitle.setproctitle(__PROC_NAME)


def server_process(cfg, debug=False):
    # describe a client
    class Bridge:
        connected: bool
        name: str
        status: dict = {}
        cli_ws = None  # cli ws sid
        web_ws_list = (
            []
        )  # frontend ws sids. client data should be able to be shown on multiple frontend

        def __init__(self, name) -> None:
            # initialize non-websocket things
            self.name = name

        def handle_ws_recv(self):
            pass

        def ws_send(self):
            pass

    # ===============================================================

    print()
    if debug:
        console.log(f"Running with debug, using APIFlask")
        from apiflask import APIFlask

        app = APIFlask(__PROC_NAME)
    else:
        console.log(f"Running in production mode, using Flask")
        from flask import Flask

        app = Flask(__PROC_NAME)

    # websocket server
    ws_server = WebsocketServer(host="0.0.0.0", port=cfg["port"] + 1)
    __BRIDGES = {}  # manage connections
    connected_clients: Dict(int, Tuple(str, str)) = {}  # {cid:(name,type)} store connection only

    # ========================  WS  SERVER  ===========================

    """ server behaviors when someone is requesting data of a Client

    !!! some axioms
    !!! a client and related frontend(s) must connect to the server with websocket
    !!! per name, per client. No duplicate names for different client
    !!! there might be multiple frontend(s) reading from single client

    frontend requesting status of name
    -   name exist: return status
    -   name not exist: return 404

    client synchronizing status of name
    -   name exist: update status
    -   name not exist:
        -   create Client with name
        -   update status


    on websocket connect:
    -   its a client with name:
        -   name exist:
            -   show warning, client websocket overwrite
            -   replace Client's ws conn sid
        -   name not exist:
            -   create Client with name (frontend may not try a ws connection before it knows the name. so not necessary to check the connection type, it must be client)
    -   its a frontend with name:
        -   name exist: add to Client's frontend ws conn list
        -   name not exist:
            -   error! frontend cannot read to a non-existing client

    on websocket disconnect:
    -   is a frontend conn
        -   remove sid from Client's frontend ws conn list
    -   is a client conn
        -   remove sid from Client
    -   remove from connected_clients dict

    on server receive json
    -   sent by client with name
        -   name exist: (name must exist)
            -   for websocket of frontend within name:
                -   forward message
    -   sent by frontend with name
        -   name exist: (name must exist)
            -   forward message to client
    """

    def handle_ws_connect(client, server):
        console.log(f"client {client} connected. waiting for handshake...")

    def handle_ws_disconnect(client, server):
        if client["id"] not in connected_clients:
            return  # client disconnected before handshake, returning anyway
        _project_name, _who = connected_clients[client["id"]]
        if _who == "cli":  # remove client from Bridge
            __BRIDGES[_project_name].cli_ws = None
        else:  # remove frontend from Bridge
            _new_web_ws_list = [
                c for c in __BRIDGES[_project_name].web_ws_list if c["id"] != client["id"]
            ]
            __BRIDGES[_project_name].web_ws_list = _new_web_ws_list
        del connected_clients[client["id"]]
        console.log(f"a {_who} disconnected with id {client['id']}")
        # logger.info(f"Websocket ({conn_type}) for {name} disconnected")

    def handle_ws_message(client, server: WebsocketServer, message):
        message_dict = json.loads(message)
        # handle event-type
        _event_type = message_dict[EVENT_TYPE_NAME_KEY]
        _payload = message_dict[PAYLOAD_NAME_KEY]
        _event_id = message_dict[EVENT_ID_NAME_KEY]
        _project_name = message_dict[NAME_NAME_KEY]
        if _event_type == "handshake":  # handle handshake
            if client["id"] in connected_clients:
                # !!! cli/web could change their project name by handshake twice, this is a legal behavior
                handle_ws_disconnect(
                    client=client, server=server
                )  # perform "software disconnect" before "software connect" again
            # assign this client to a Bridge
            _who = _payload["who"]
            console.log(f"handling handshake for {_who} with name {_project_name}")
            if _who == "web":
                # new connection from frontend
                # check if Bridge with name exist
                if _project_name not in __BRIDGES:  # there is no such bridge
                    server.send_message(
                        client=client,
                        msg=json.dumps(
                            WsMsg(
                                name=_project_name,
                                event_type="handshake",
                                event_id=_event_id,
                                payload={"result": 404, "reason": "name not found"},
                            ).json()
                        ),
                    )
                else:  # assign web to bridge
                    _target_bridge = __BRIDGES[_project_name]
                    _target_bridge.web_ws_list.append(client)
                    connected_clients[client["id"]] = (_project_name, "web")
                    server.send_message(
                        client=client,
                        msg=json.dumps(
                            WsMsg(
                                name=_project_name,
                                event_type="handshake",
                                event_id=_event_id,
                                payload={"result": 200, "reason": "join success"},
                            ).json()
                        ),
                    )
            elif _who == "cli":
                # new connection from cli
                # check if Bridge with name exist
                if _project_name not in __BRIDGES:  # there is no such bridge
                    _target_bridge = Bridge(name=_project_name)  # create new bridge for this name
                    __BRIDGES[_project_name] = _target_bridge
                __BRIDGES[_project_name].cli_ws = client  # assign cli to bridge
                connected_clients[client["id"]] = (_project_name, "cli")
                server.send_message(
                    client=client,
                    msg=json.dumps(
                        WsMsg(
                            name=_project_name,
                            event_type="handshake",
                            event_id=_event_id,
                            payload={"result": 200, "reason": "join success"},
                        ).json()
                    ),
                )
            return  # return after handling handshake

        if client["id"] not in connected_clients:
            return  # !!! not handling messages from cli/web without handshake. handshake is a special case and should be handled anyway before this check.

        _, _who = connected_clients[client["id"]]  # check if is web or cli

        def send_to_frontends_of_name(name, message):
            if name not in __BRIDGES:
                if debug:
                    console.log(
                        f"cannot broadcast message to frontends under name {name}: name not found."
                    )
                return  # no such bridge
            _target_bridge = __BRIDGES[name]
            for web_ws in _target_bridge.web_ws_list:
                server.send_message(
                    client=web_ws, msg=message
                )  # forward original message to frontend
            return

        def send_to_client_of_name(name, message):
            if name not in __BRIDGES:
                if debug:
                    console.log(
                        f"cannot forward message to client under name {name}: name not found."
                    )
                return  # no such bridge
            _target_bridge = __BRIDGES[name]
            _client = _target_bridge.cli_ws
            server.send_message(client=_client, msg=message)  # forward original message to client
            return

        if _event_type == "log":  # handle log
            # forward log to frontend. logs should only be sent by cli and only be received by frontends
            if _project_name not in __BRIDGES:
                # project name must exist
                # drop anyway if not exist
                if debug:
                    console.log(f"handle log. {_project_name} not found.")
                return
            else:
                send_to_frontends_of_name(
                    name=_project_name, message=message
                )  # forward to frontends
            return  # return after handling log forwarding

        if _event_type == "action":
            if _who == "web":  # frontend send action query to client
                send_to_client_of_name(_project_name, message=message)
            else:  # _who == 'cli', client send action result back to frontend(s)
                send_to_frontends_of_name(_project_name, message=message)
            return  # return after handling action forwarding

        if _event_type == "ack":
            # todo forward ack to waiting acks
            return  # return after handling ack

    ws_server.set_fn_new_client(handle_ws_connect)
    ws_server.set_fn_client_left(handle_ws_disconnect)
    ws_server.set_fn_message_received(handle_ws_message)

    # ======================== HTTP  SERVER ===========================

    # @app.route(f"{FRONTEND_API_ROOT}/wsforward/<name>", methods=["POST"])
    # def handle_json_forward_to_client_ws(name):  # forward frontend http json to client ws
    #     data = request.json
    #     if name in __BRIDGES:  # client name exist
    #         target_sid = __BRIDGES[name].cli_ws
    #         if target_sid is None:  # no active cli ws connection for this name
    #             # logger.warn(
    #             #     f"frontend tried to talk to forward to a disconnected client ws with name {name}."
    #             # )
    #             abort(404)
    #         ws_send(data, to=target_sid)
    #         return "ok"

    front_end_dist_path = os.path.join(os.path.dirname(__file__), "../../frontend_dist")

    @app.route("/")
    def static_serve_root():
        return send_from_directory(front_end_dist_path, "index.html")

    @app.route("/<path:name>")
    def static_serve(name):
        try:
            return send_from_directory(front_end_dist_path, name)
        except werkzeug.exceptions.NotFound:
            return send_from_directory(front_end_dist_path, "index.html")

    @app.route("/hello", methods=["GET"])
    def just_send_hello():
        return json.dumps({"hello": "hello"})

    @app.route(f"{FRONTEND_API_ROOT}/status/<name>", methods=["GET"])
    def return_status_of(name):
        if name in __BRIDGES:
            _returning_stat = __BRIDGES[name].status  # returning specific status
        else:
            abort(404)
        return _returning_stat

    @app.route(f"{FRONTEND_API_ROOT}/list", methods=["GET"])
    def return_names_of_status():
        _names = {"names": list(__BRIDGES.keys())}
        return _names

    @app.route(f"{CLIENT_API_ROOT}/sync/<name>", methods=["POST"])
    def sync_status_of(name):  # client side function
        _json_data = request.get_json()
        if name not in __BRIDGES:  # Client not found
            __BRIDGES[name] = Bridge(name=name)  # Create from sync request
        __BRIDGES[name].status = _json_data

        return "ok"

    @app.route(f"{FRONTEND_API_ROOT}/shutdown", methods=["POST"])
    def shutdown():
        def __sleep_and_shutdown(secs=3):
            time.sleep(secs)
            os._exit(0)

        Thread(target=__sleep_and_shutdown).start()  # shutdown after 3 seconds
        console.log(f"BYE.")
        return f"shutdown in {3} seconds."

    console.log(f"launching websocket server...")
    ws_server.run_forever(threaded=True)

    _port = cfg["port"]
    console.log(f"visit server at http://localhost:{_port}")
    app.run(host="0.0.0.0", port=_port)


if __name__ == "__main__":
    cfg = {
        "enable": True,
        "host": "localhost",
        "port": 5000,
        "displayName": None,
        "allowIpython": False,
        "mute": True,
        "mode": "detached",
        "uploadInterval": 10,
    }
    server_process(cfg, debug=True)



import json
import os

import click

import neetbox
import neetbox.daemon.server as server_module
from neetbox.config import get_module_level_config
from neetbox.daemon.client._client_apis import *
from neetbox.daemon.server._server import server_process
from neetbox.logging.formatting import LogStyle
from neetbox.logging.logger import Logger

logger = Logger("NEETBOX", style=LogStyle(with_datetime=False, skip_writers=["ws"]))


@click.group()
@click.option(
    "--verbose",
    "-v",
    help="set verbose flag to True",
    default=False,
)
@click.pass_context
def main(ctx, verbose: bool):
    ctx.ensure_object(dict)
    """This is NEETBOX cli"""
    ctx.obj["verbose"] = verbose


def _try_load_workspace_if_applicable():
    if neetbox._is_in_workspace():
        success = neetbox._load_workspace()
        if not success:
            os._exit(255)


@main.command(name="list")
def list_command():
    """Show list of connected project names"""
    _try_load_workspace_if_applicable()
    try:
        _list_json = get_list()
        click.echo(json.dumps(_list_json))
    except Exception as e:  # noqa
        logger.log("Could not fetch data. Is there any project with NEETBOX running?")


@main.command(name="status")
@click.argument("name")
def status_command(name):
    """Show the working tree status"""
    _try_load_workspace_if_applicable()
    _stat_json = None
    try:
        _stat_json = get_status_of(name)
        click.echo(json.dumps(_stat_json))
    except Exception as e:  # noqa
        logger.log("Could not fetch data. Is there any project with NEETBOX running?")


@main.command(name="serve")
@click.option("--port", "-p", help="specify which port to launch", metavar="port", required=False)
def serve(port: int):
    """serve neetbox server in attached mode"""
    _try_load_workspace_if_applicable()
    try:
        server_config = get_module_level_config(server_module)
        logger.log(f"Launching server using config: {server_config}")
        server_process(cfg=server_config)
    except Exception as e:
        logger.err(f"Failed to launch a neetbox server: {e}")


@main.command()
@click.option("--name", "-n", help="set project name", metavar="name", required=False)
def init(name: str):
    """initialize current folder as workspace and generate the config file from defaults"""
    try:
        if neetbox._init_workspace(name=name):
            logger.console_banner("neetbox", font="ansishadow")
            logger.log("Welcome to NEETBOX")
    except Exception as e:
        logger.err(f"Failed to init here: {e}")


if __name__ == "__main__":
    main()



import argparse
import json
import sys

from neetbox.daemon.server._server import server_process

# sys.stdout=open(r'D:\Projects\ML\neetbox\logdir\daemon.log', 'a+')

print("========= Daemon  =========")


def run():
    if len(sys.argv) <= 1:
        print("_daemon_: Warning: empty daemon_config")
        daemon_config = None
    else:
        ap = argparse.ArgumentParser()
        ap.add_argument("--config")
        args = ap.parse_args()
        daemon_config = json.loads(args.config)
        print("Daemon started with config:", daemon_config)
    server_process(daemon_config)


print("_daemon_ is starting with __name__ =", __name__, " and args =", sys.argv)
run()
print("_daemon_: exiting")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230318

from dataclasses import dataclass
from random import random
from typing import Optional


@dataclass
class LogStyle:
    console_color: Optional[str] = None
    prefix: str = ""
    text_style: Optional[str] = None
    datetime_format: str = "%Y-%m-%d-%H:%M:%S"
    with_identifier: bool = True
    trace_level = 3
    with_datetime: bool = True
    split_char_cmd = " > "
    split_char_identity = "/"
    split_char_txt = " | "
    skip_writers: list[str] = None

    @classmethod
    def get_supported_colors(cls):
        return ["red", "green", "blue", "cyan", "yellow", "magenta"]

    @classmethod
    def get_supported_text_style(cls):
        return ["bold", "italic", "blink"]

    def parse(self, pattern: str):
        # todo
        pass

    def set_color(self, color: str):
        self.fore = color
        return self

    def set_text_style(self, text_style: str):
        self.text_style = text_style
        return self

    def set_prefix(self, prefix: str):
        self.prefix = prefix
        return self

    def set_datetime_format(self, datetime_format: str):
        self.datetime_format = datetime_format
        return self

    def randcolor(self):
        colors = LogStyle.get_supported_colors()
        # split_index = int(random() * len(colors) / 2)
        # index_offset = -1
        # while index_offset == 0:  # fore and back shoud not be the same
        #     index_offset = int(random() * len(colors) / 2)
        # self.back = colors[(split_index + index_offset) % len(colors)]
        # self.fore = colors[(split_index - index_offset) % len(colors)]
        self.console_color = colors[int(random() * len(colors))]
        return self


DEFAULT_STYLE = LogStyle()


def styled_text(text, style: LogStyle):
    attributes = []
    if style.console_color:
        attributes.append(style.console_color)
    if style.text_style:
        attributes.append(style.text_style)
    render_stack = []
    _prefix = ""
    _postfix = ""
    while len(attributes):
        attr = attributes.pop(-1)
        _prefix += f"[{attr}]"
        render_stack.append(attr)
    while len(render_stack):
        _postfix += f"[/{render_stack.pop(-1)}]"
    return f"{_prefix}{text}{_postfix}"


def colored_text(text: str, color):
    return f"[{color}]{text}[/{color}]"



import os
import sys

import toml

import neetbox.daemon as daemon
import neetbox.integrations as integrations
from neetbox.config import default as default_config
from neetbox.config._config import update_workspace_config_with
from neetbox.logging.formatting import LogStyle
from neetbox.logging.logger import Logger
from neetbox.utils.framing import get_frame_module_traceback

logger = Logger("NEETBOX", style=LogStyle(with_datetime=False, skip_writers=["ws"]))

MODULE_NAME = get_frame_module_traceback(1).__name__  # type: ignore
CONFIG_FILE_NAME = f"{MODULE_NAME}.toml"


def _init_workspace(path=None, **kwargs) -> bool:
    if path:
        os.chdir(path=path)
    current_path = os.getcwd()
    config_file_path = os.path.join(current_path, CONFIG_FILE_NAME)
    if not os.path.isfile(config_file_path):  # config file not exist
        try:  # creating config file using default config
            with open(config_file_path, "w+") as config_file:
                _config = dict(default_config)
                if "name" in kwargs and kwargs["name"]:  # using given name
                    _config["name"] = kwargs["name"]
                else:  # using the folder name
                    _config["name"] = os.path.basename(os.path.normpath(os.getcwd()))
                toml.dump(_config, config_file)
            logger.ok(f"Workspace config created as {config_file_path}.")
            return True
        except Exception as e:
            logger.err(f"Failed to create {config_file_path}: {e}")
            return False
    else:  # config file exist:
        logger.err((f"Config file already exists."))  # noqa
        return False


def _load_workspace(path=None) -> bool:
    if path:
        os.chdir(path=path)
    current_path = os.getcwd()
    config_file_path = os.path.join(current_path, CONFIG_FILE_NAME)
    if not os.path.isfile(config_file_path):  # but config file not exist
        logger.err(f"Config file {config_file_path} not exists.")
        return False
    try:  # do load
        load_config = toml.load(config_file_path)
        update_workspace_config_with(load_config)
        logger.ok(f"workspace config loaded from {config_file_path}.")
        return True
    except Exception as e:
        logger.err(f"failed to load workspace config from {config_file_path}: {e}")
        return False


is_in_daemon_process = (
    "NEETBOX_DAEMON_PROCESS" in os.environ and os.environ["NEETBOX_DAEMON_PROCESS"] == "1"
)


def _load_workspace_as_a_project():
    success = _load_workspace()  # init from config file
    if not success:  # failed to load workspace config, exiting
        os._exit(255)
    # post init
    integrations._post_init_workspace()
    daemon._try_attach_daemon()


def _is_in_workspace():
    return os.path.isfile(CONFIG_FILE_NAME)


if len(sys.argv) > 0 and sys.argv[0].endswith("neet") or is_in_daemon_process:
    # running in cli or daemon process, do not load workspace
    pass
elif _is_in_workspace():  # if a config file exist and not running in cli
    _load_workspace_as_a_project()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230413

import importlib
from enum import Enum
from functools import lru_cache
from typing import List, Optional

from neetbox.logging import logger


class Engine(Enum):
    Torch = "torch"

    def __str__(self) -> str:
        return self.value


supported_engines: Optional[List] = None
installed_engines: Optional[List] = None


@lru_cache
def get_supported_engines():
    global supported_engines
    if not supported_engines:
        supported_engines = []
        for engine in Engine:
            supported_engines.append(engine)
    return supported_engines.copy()


def get_installed_engines():
    global installed_engines
    if not installed_engines:
        logger.info("Scanning installed engines...")
        installed_engines = []
        for engine in get_supported_engines():
            try:
                importlib.import_module(engine.value)
                installed_engines.append(engine)
                logger.info(f"'{engine.vaule}' was found installed.")
            except ImportError:
                pass
    return installed_engines.copy()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230414

import json
import time
from threading import Thread
from typing import Union

from neetbox.config import get_module_level_config
from neetbox.daemon.client._client import connection
from neetbox.daemon.server._server import CLIENT_API_ROOT
from neetbox.logging.formatting import LogStyle
from neetbox.logging.logger import Logger
from neetbox.pipeline._signal_and_slot import _UPDATE_VALUE_DICT, SYSTEM_CHANNEL, watch

logger = Logger(style=LogStyle(with_datetime=False, skip_writers=["ws"]))


def _add_upload_thread_to_watch(daemon_config, base_addr, display_name):
    _api_name = "sync"
    _api_addr = f"{base_addr}{CLIENT_API_ROOT}/{_api_name}/{display_name}"

    @watch(interval=daemon_config["uploadInterval"], overwrite=True)
    def upload_via_http():
        # dump status as json
        _data = json.dumps(_UPDATE_VALUE_DICT[SYSTEM_CHANNEL], default=str)
        _headers = {"Content-Type": "application/json"}
        try:
            # upload data
            resp = connection.http.post(_api_addr, json=_data, headers=_headers)
            if resp.is_error:  # upload failed
                raise IOError(f"Failed to upload data to daemon. ({resp.status_code})")
        except Exception as e:
            logger.warn(f"Failed to upload data to daemon cause {e}. Waiting for reconnect...")


def connect_daemon(cfg=None, launch_upload_thread=True):
    cfg = cfg or get_module_level_config()
    _display_name = get_module_level_config()["displayName"]
    _launch_config = get_module_level_config("@")
    _display_name = _display_name or _launch_config["name"]

    logger.log(f"Connecting to daemon at {cfg['host']}:{cfg['port']} ...")
    _daemon_server_address = f"{cfg['host']}:{cfg['port']}"
    _base_addr = f"http://{_daemon_server_address}"

    # check if daemon is alive
    def _check_daemon_alive(_api_addr):
        _api_name = "hello"
        _api_addr = f"{_api_addr}/{_api_name}"
        r = connection.http.get(_api_addr)
        if r.is_error:
            raise IOError(f"Daemon at {_api_addr} is not alive. ({r.status_code})")

    try:
        _check_daemon_alive(_base_addr)
        logger.ok(f"daemon alive at {_base_addr}")
    except Exception as e:
        logger.err(e)
        return False

    if launch_upload_thread:
        _add_upload_thread_to_watch(
            daemon_config=cfg, base_addr=_base_addr, display_name=_display_name
        )

    return True



import inspect
import io
import json
import os
import pathlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Callable, Iterable, Optional, Union

from rich import print as rprint

from neetbox.logging.formatting import LogStyle, colored_text, styled_text
from neetbox.utils import formatting


class LogWriter(ABC):
    @abstractmethod
    def write(self, raw_log):
        pass


# ================== DEFINE LOG TYPE =====================


@dataclass
class RawLog:
    rich_msg: str
    style: LogStyle
    caller_identity: Any
    whom: Any = None
    prefix: Optional[str] = None
    datetime_format: Optional[str] = None
    with_identifier: Optional[bool] = None
    with_datetime: Optional[bool] = None
    skip_writers: Optional[list[str]] = None

    def write_by(self, writer: LogWriter) -> bool:
        _skip_writers = (self.style.skip_writers or []) + (self.skip_writers or [])
        for swr in _skip_writers:
            if isinstance(writer, RawLog.name2writerType[swr]):
                return False  # skip this writer, do not write
        writer.write(self)
        return False

    def json(self) -> dict:
        _default_style = self.style
        # prefix
        _prefix = self.prefix or _default_style.prefix
        # composing datetime
        _with_datetime = (
            _default_style.with_datetime if self.with_datetime is None else self.with_datetime
        )
        _datetime = ""
        if _with_datetime:
            _datetime_fmt = self.datetime_format or _default_style.datetime_format
            _datetime = datetime.now().strftime(_datetime_fmt)

        # composing identifier
        _whom = ""
        _with_identifier = (
            _default_style.with_identifier if self.with_identifier is None else self.with_identifier
        )
        if _with_identifier:
            _caller_identity = self.caller_identity
            _whom = str(self.whom)  # check identity
            id_seq = []
            if self.whom is None:  # if using default logger, tracing back to the caller
                file_level = True
                _whom = ""
                if _caller_identity.module_name and _default_style.trace_level >= 2:
                    # trace as module level
                    id_seq.append(_caller_identity.module_name)
                    file_level = False
                if _caller_identity.class_name and _default_style.trace_level >= 1:
                    # trace as class level
                    id_seq.append(_caller_identity.class_name)
                    file_level = False
                if file_level and _default_style.trace_level >= 1:
                    id_seq.append(_caller_identity.filename)  # not module level and class level
            if _caller_identity.func_name != "<module>":
                id_seq.append(_caller_identity.func_name)  # skip for jupyters
            for i in range(len(id_seq)):
                if len(_whom) != 0:
                    _whom += _default_style.split_char_identity
                _whom += id_seq[i]
        return {"prefix": _prefix, "datetime": _datetime, "whom": _whom, "msg": self.rich_msg}

    def __repr__(self) -> str:
        return json.dumps(self.json(), default=str)


# ================== CONSOLE LOG WRITER =====================


class __ConsoleLogWriter(LogWriter):
    def write(self, raw_log: RawLog):
        _msg_dict = raw_log.json()
        _style = raw_log.style
        rich_msg = str(
            _msg_dict["prefix"]
            + _msg_dict["datetime"]
            + _style.split_char_cmd * min(len(_msg_dict["datetime"]), 1)
            + styled_text(_msg_dict["whom"], style=_style)
            + _style.split_char_cmd * min(len(_msg_dict["whom"]), 1)
            + _msg_dict["msg"]
        )
        rprint(rich_msg)


# console writer singleton
consoleLogWriter = __ConsoleLogWriter()


# ================== FILE LOG WRITER =====================


class FileLogWriter(LogWriter):
    # class level static pool
    PATH_2_FILE_WRITER = {}

    # instance level non-static things
    file_writer = None  # assign in __init__

    def __new__(cls, path, *args, **kwargs):
        # per file, per writer.
        file_abs_path = os.path.abspath(path)
        if os.path.isdir(file_abs_path):
            raise Exception("Target path is not a file.")
        filename = formatting.legal_file_name_of(os.path.basename(path))
        dirname = os.path.dirname(path) if len(os.path.dirname(path)) != 0 else "."
        if not os.path.exists(dirname):
            raise Exception(f"Could not find dictionary {dirname}")
        real_path = os.path.join(dirname, filename)
        if file_abs_path not in FileLogWriter.PATH_2_FILE_WRITER:
            newWriter = LogWriter.__new__(cls)
            newWriter.file_writer = open(real_path, "a", encoding="utf-8", buffering=1)
            FileLogWriter.PATH_2_FILE_WRITER[file_abs_path] = newWriter

        return FileLogWriter.PATH_2_FILE_WRITER[file_abs_path]

    def __init__(self, path) -> None:
        self.file_writer = open(path, "a", encoding="utf-8", buffering=1)

    def write(self, raw_log: RawLog):
        _msg_dict = raw_log.json()
        _style = raw_log.style
        text_msg = str(
            _msg_dict["prefix"]
            + _msg_dict["datetime"]
            + _style.split_char_txt * min(len(_msg_dict["datetime"]), 1)
            + _msg_dict["whom"]
            + _style.split_char_txt * min(len(_msg_dict["whom"]), 1)
            + _msg_dict["msg"]
            + "\n"
        )
        self.file_writer.write(text_msg)


# ================== WS LOG WRITER =====================


class _WebSocketLogWriter(LogWriter):
    # class level statics
    connection = None  # connection should be assigned by neetbox.daemon.client._client to avoid recursive import

    def write(self, raw_log: RawLog):
        json_data = raw_log.json()

        if _WebSocketLogWriter.connection:
            _WebSocketLogWriter.connection.ws_send(event_type="log", payload=json_data)


def _assign_connection_to_WebSocketLogWriter(conn):
    _WebSocketLogWriter.connection = conn


webSocketLogWriter = _WebSocketLogWriter()


# ================== JSON LOG WRITER =====================


class JsonLogWriter(FileLogWriter):
    def write(self, raw_log: RawLog):
        # todo convert to json and write to file
        pass


# ================== POST INIT TYPE REF =====================

RawLog.name2writerType = {
    "stdout": __ConsoleLogWriter,
    "file": FileLogWriter,
    "ws": _WebSocketLogWriter,
    "json": JsonLogWriter,
}



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230417

import getpass
import locale
import platform
import subprocess

from neetbox.config import get_module_level_config
from neetbox.integrations import call_on_workspace_load
from neetbox.pipeline import watch
from neetbox.pipeline._signal_and_slot import SYSTEM_CHANNEL
from neetbox.utils.mvc import Singleton


class __Platform(dict, metaclass=Singleton):
    def __init__(self):
        # system
        self["username"] = getpass.getuser()
        self["machine"] = platform.machine()
        self["processor"] = "unknown" if len(platform.processor()) == 0 else platform.processor()
        self["os_name"] = platform.system()
        self["os_release"] = platform.version()
        self["architecture"] = platform.architecture()
        self["python_version"] = platform.python_version()
        self["python_build"] = platform.python_build()

    def exec(self, command):
        """
        Run a terminal command.

        Args:
            command (str): The command need to run.

        Returns:
            int: The command return code.
            str: The command running results.
            err: The command error information.
        """
        p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        raw_output, raw_err = p.communicate()
        rc = p.returncode
        if self.platform_info["architecture"] == "32bit":
            enc = "oem"
        else:
            enc = locale.getpreferredencoding()
        output = raw_output.decode(enc)
        err = raw_err.decode(enc)

        return rc, output.strip(), err.strip()


# singleton
platform = __Platform()


# watch updates in daemon
@call_on_workspace_load(name="show-platform-information")
def load_monit_hardware():
    cfg = get_module_level_config()
    if cfg["monit"]:  # if do monit hardware
        # watch updates in daemon
        @watch(name="platform", initiative=True, _channel=SYSTEM_CHANNEL)
        def update_env_stat():
            return dict(platform)

        update_env_stat()  # call once



import functools
import json
import logging
from collections import defaultdict
from threading import Thread
from typing import Callable

import httpx
import websocket

from neetbox.config import get_module_level_config
from neetbox.daemon._protocol import *
from neetbox.logging.formatting import LogStyle
from neetbox.logging.logger import Logger
from neetbox.utils.mvc import Singleton

logger = Logger(whom=None, style=LogStyle(skip_writers=["ws"]))

httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.ERROR)

_ws_initialized = False  # indicate whether websocket has been initialized


def _load_http_client():
    __local_http_client = httpx.Client(
        proxies={
            "http://": None,
            "https://": None,
        }
    )  # type: ignore
    return __local_http_client


# singleton
class ClientConn(metaclass=Singleton):
    # http client
    http: httpx.Client = _load_http_client()

    __ws_client: websocket.WebSocketApp = None  # _websocket_client
    __ws_subscription = defaultdict(lambda: {})  # default to no subscribers

    def ws_subscribe(event_type_name: str, name: str = None):
        """let a function subscribe to ws messages with event type name.
        !!! dfor inner APIs only, do not use this in your code!
        !!! developers should contorl blocking on their own functions

        Args:
            function (Callable): who is subscribing the event type
            event_type_name (str, optional): Which event to listen. Defaults to None.
        """
        return functools.partial(
            ClientConn._ws_subscribe, event_type_name=event_type_name, name=name
        )

    def _ws_subscribe(function: Callable, event_type_name: str, name=None):
        name = name or function.__name__
        ClientConn.__ws_subscription[event_type_name][name] = function
        logger.debug(f"ws: {name} subscribed to '{event_type_name}")

    def _init_ws():
        global _ws_initialized
        if _ws_initialized:
            return

        cfg = get_module_level_config()
        _root_config = get_module_level_config("@")
        ClientConn._display_name = cfg["displayName"] or _root_config["name"]

        # ws server url
        ClientConn.ws_server_addr = f"ws://{cfg['host']}:{cfg['port'] + 1}"

        # create websocket app
        logger.log(
            f"creating websocket connection to {ClientConn.ws_server_addr}", skip_writers=["ws"]
        )
        ClientConn.wsApp = websocket.WebSocketApp(
            ClientConn.ws_server_addr,
            on_open=ClientConn.__on_ws_open,
            on_message=ClientConn.__on_ws_message,
            on_error=ClientConn.__on_ws_err,
            on_close=ClientConn.__on_ws_close,
        )

        Thread(
            target=ClientConn.wsApp.run_forever, kwargs={"reconnect": True}, daemon=True
        ).start()  # initialize and start ws thread

        _ws_initialized = True

    def __on_ws_open(ws: websocket.WebSocketApp):
        _display_name = ClientConn._display_name
        logger.ok(f"client websocket connected. sending handshake as '{_display_name}'...")
        ws.send(  # send handshake request
            json.dumps(
                {
                    NAME_NAME_KEY: _display_name,
                    EVENT_TYPE_NAME_KEY: "handshake",
                    PAYLOAD_NAME_KEY: {"who": "cli"},
                    EVENT_ID_NAME_KEY: 0,  # todo how does ack work
                },
                default=str,
            )
        )

        @ClientConn.ws_subscribe(event_type_name="handshake")
        def _handle_handshake(msg):
            assert msg[PAYLOAD_NAME_KEY]["result"] == 200
            logger.ok(f"handshake succeed.")
            ClientConn.__ws_client = ws

    def __on_ws_err(ws: websocket.WebSocketApp, msg):
        logger.err(f"client websocket encountered {msg}")

    def __on_ws_close(ws: websocket.WebSocketApp, close_status_code, close_msg):
        logger.warn(f"client websocket closed")
        if close_status_code or close_msg:
            logger.warn(f"ws close status code: {close_status_code}")
            logger.warn("ws close message: {close_msg}")
        ClientConn.__ws_client = None

    def __on_ws_message(ws: websocket.WebSocketApp, msg):
        """EXAMPLE JSON
        {
            "event-type": "action",
            "event-id": 111 (optional?)
            "payload": ...
        }
        """
        msg = json.loads(msg)  # message should be json
        logger.debug(f"ws received {msg}")

        event_type_name = msg[EVENT_TYPE_NAME_KEY]
        if event_type_name not in ClientConn.__ws_subscription:
            logger.warn(
                f"Client received a(n) {event_type_name} event but nobody subscribes it. Ignoring anyway."
            )
        for name, subscriber in ClientConn.__ws_subscription[event_type_name].items():
            try:
                subscriber(msg)  # pass payload message into subscriber
            except Exception as e:
                # subscriber throws error
                logger.err(
                    f"Subscriber {name} crashed on message event {event_type_name}, ignoring."
                )

    def ws_send(event_type: str, payload, event_id=-1):
        if ClientConn.__ws_client:  # if ws client exist
            ClientConn.__ws_client.send(
                json.dumps(
                    {
                        NAME_NAME_KEY: ClientConn._display_name,
                        EVENT_TYPE_NAME_KEY: event_type,
                        PAYLOAD_NAME_KEY: payload,
                        EVENT_ID_NAME_KEY: event_id,
                    },
                    default=str,
                )
            )
        else:
            if not _ws_initialized:  # ws not intialized
                ClientConn._init_ws()


# assign this connection to websocket log writer
from neetbox.logging._writer import _assign_connection_to_WebSocketLogWriter

_assign_connection_to_WebSocketLogWriter(ClientConn)
connection = ClientConn



import matplotlib.pyplot as plt
from matplotlib import ticker


def draw_split_y_plot(x, y, split_range=[(0, 5), (10, 30)]):
    """
    画 y 轴的断开图线
    """
    """
    问题较多，有空再修，和 x 轴差不多
    """
    # If we were to simply plot data, we'd lose most of the interesting
    # details due to the outliers. So let's 'break' or 'cut-out' the y-axis
    # into two portions - use the top (ax1) for the outliers, and the bottom
    # (ax2) for the details of the majority of our data
    fig, (ax1, ax2, ax3) = plt.subplots(2, 1, sharex=True)
    fig.subplots_adjust(hspace=0.05)  # adjust space between axes

    # plot the same data on both axes
    ax1.plot(x, y)
    ax2.plot(x, y)

    # zoom-in / limit the view to different portions of the data
    part1, part2 = split_range
    ax1.set_xlim(part1[0], part1[1])  # outliers only
    ax2.set_xlim(part2[0], part2[1])  # most of the data

    # hide the spines between ax and ax2
    ax1.spines.bottom.set_visible(False)
    ax2.spines.top.set_visible(False)
    ax1.xaxis.tick_top()
    ax1.tick_params(labeltop=False)  # don't put tick labels at the top
    ax2.xaxis.tick_bottom()

    # Now, let's turn towards the cut-out slanted lines.
    # We create line objects in axes coordinates, in which (0,0), (0,1),
    # (1,0), and (1,1) are the four corners of the axes.
    # The slanted lines themselves are markers at those locations, such that the
    # lines keep their angle and position, independent of the axes size or scale
    # Finally, we need to disable clipping.

    # draw split line
    d = 0.5  # proportion of vertical to horizontal extent of the slanted line
    kwargs = dict(
        marker=[(-1, -d), (1, d)],
        markersize=12,
        linestyle="none",
        color="k",
        mec="k",
        mew=1,
        clip_on=False,
    )
    ax1.plot([0, 1], [0, 0], transform=ax1.transAxes, **kwargs)
    ax2.plot([0, 1], [1, 1], transform=ax2.transAxes, **kwargs)
    ax3.plot([1, 0], [0, 1], transform=ax2.transAxes, **kwargs)

    plt.show()


def draw_split_x_plot(x, y, split_range=[(0, 5), (10, 30), (50, 100)], is_show=False):
    """
    画 x 轴的断开图线
    """
    """
    todo: 1. 加入设置 title 的功能
    todo: 2. 提供输入分块数量，根据分块数量自动分块
    todo: 3. 优化设置分块范围
    todo: 4. 加入设置横纵坐标的 label
    todo: 5. 加入设置非数值型刻度
    todo: 6. 根据数据的长度自动适应横坐标刻度间隔
    """
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True)
    fig.subplots_adjust(wspace=0.06)  # adjust space between axes

    # plot the same data on both axes
    ax1.plot(x, y)
    ax2.plot(x, y)
    ax3.plot(x, y)

    # zoom-in / limit the view to different portions of the data
    part1, part2, part3 = split_range
    ax1.set_xlim(part1[0], part1[1])  # outliers only
    ax2.set_xlim(part2[0], part2[1])  # most of the data
    ax3.set_xlim(part3[0], part3[1])  # most of the data

    # hide the spines between ax and ax2
    ax1.spines["right"].set_visible(False)
    ax2.spines["left"].set_visible(False)
    ax2.spines["right"].set_visible(False)
    ax3.spines["left"].set_visible(False)
    ax1.yaxis.tick_left()
    ax2.yaxis.tick_right()
    ax2.tick_params(right=False)
    ax3.tick_params(left=False)

    # draw split line
    d = 0.5  # proportion of vertical to horizontal extent of the slanted line
    kwargs = dict(
        marker=[(-1, -d), (1, d)],
        markersize=12,
        linestyle="none",
        color="k",
        mec="k",
        mew=1,
        clip_on=False,
    )
    ax1.plot([1, 1], [1, 0], transform=ax1.transAxes, **kwargs)
    ax2.plot([0, 0], [1, 0], transform=ax2.transAxes, **kwargs)
    ax2.plot([1, 1], [1, 0], transform=ax2.transAxes, **kwargs)
    ax3.plot([0, 0], [1, 0], transform=ax3.transAxes, **kwargs)

    fig.set_figheight(5)
    fig.set_figwidth(10)
    fig.text(0.5, -0.16, "Subject Object Pair", ha="center", fontsize=14)

    ax1.set_ylabel("mR", fontsize=14)
    x_list = list(range(len(y)))
    ax1.set_xticks(x_list[part1[0] : part1[1]], x[part1[0] : part1[1]], rotation=60)
    ax2.set_xticks(x_list[part2[0] : part2[1]], x[part2[0] : part2[1]], rotation=60)
    ax3.set_xticks(x_list[part3[0] : part3[1]], x[part3[0] : part3[1]], rotation=60)
    ax1.xaxis.set_major_locator(ticker.MultipleLocator(8))
    ax2.xaxis.set_major_locator(ticker.MultipleLocator(22))
    ax3.xaxis.set_major_locator(ticker.MultipleLocator(8))

    if is_show:
        plt.show()

    return fig, ax1, ax2, ax3


def sorted_dict(data, key=0, up_or_down=True):
    """
    sorted dict by key
    Args:
        data (dict): a data dict
        key (int, optional): choose sorted by key or value, 0 means sorted by key, 1 means value. Defaults to 0.
        up_or_down (bool, optional): Whether reverse. Defaults to True.

    Returns:
        dict: sorted dict
    """
    return dict(sorted(data.items(), key=lambda x: x[key], reverse=up_or_down))



from typing import Any, Dict, List

import uvicorn
from fastapi import FastAPI, HTTPException, WebSocket
from pydantic import BaseModel
from starlette.endpoints import WebSocketEndpoint

FRONTEND_API_ROOT = "/web"
CLIENT_API_ROOT = "/cli"

app = FastAPI()


# ===============================================================
# Client functions (backend)
# ===============================================================


class ClientEndpoint(WebSocketEndpoint):
    encoding = "json"
    subscriptions: Dict[str, List[WebSocket]] = {}
    socket_pool: Dict[str, WebSocket] = {}

    def __init__(self, scope, receive, send, name: str):
        super().__init__(scope, receive, send)
        self.name = name

    async def on_connect(self, websocket: WebSocket):
        await websocket.accept()
        ClientEndpoint.socket_pool[self.name] = websocket  # add to socket pool

    async def on_receive(self, websocket: WebSocket, data: Any):
        """
                    ┌────►Viewer
                    │
                    │
        Client─────►Center──►Viewer
                    │
                    │
                    └────►Viewer
        """
        for ws in ClientEndpoint.subscriptions[self.name]:
            await ws.send_json(data)

    async def on_disconnect(self, websocket: WebSocket, close_code: int):
        pass

    @staticmethod
    def subscribe(websocket: WebSocket, name: str):
        if name not in ClientEndpoint.subscriptions.keys():
            ClientEndpoint.subscriptions[name] = []
        ClientEndpoint.subscriptions[name].append(websocket)


@app.websocket(f"{CLIENT_API_ROOT}" + "/ws/")
async def handle_client_websocket(websocket: WebSocket, name: str):
    await ClientEndpoint(websocket.scope, websocket.receive, websocket.send, name)


class Client(BaseModel):
    status: dict = {}


class ClientManager:
    def __init__(self) -> None:
        self._client_registry: Dict[str, Client] = {}

    def register(self, name: str, client: Client):
        if name in self._client_registry.keys():
            raise ValueError(f"Client with name {name} already exists.")
        self._client_registry[name] = client

    def get(self, name: str):
        return self._client_registry[name]

    def get_all(self):
        return self._client_registry


client_manager = ClientManager()


@app.get("/register/")
async def register_client(name: str):
    try:
        client_manager.register(name, Client())
    except ValueError:
        raise HTTPException(status_code=400, detail="Client already exists.")


class Status(BaseModel):
    status: dict = {}


@app.post(f"{CLIENT_API_ROOT}" + "/sync/")
async def sync_client(name: str, status: Status):
    try:
        client = client_manager.get(name)
    except ValueError:
        raise HTTPException(status_code=404, detail="Client not found.")
    client.status = status.status


# ===============================================================
# Viewer functions (frontend)
# ===============================================================


class ViewerEndpoint(WebSocketEndpoint):
    encoding = "json"

    def __init__(self, scope, receive, send, name: str):
        super().__init__(scope, receive, send)
        self.name = name

    async def on_connect(self, websocket: WebSocket):
        await websocket.accept()
        ClientEndpoint.subscribe(websocket, self.name)

    async def on_receive(self, websocket: WebSocket, data: Any):
        """
        Viewer─────►Center─────►Client
        """
        await ClientEndpoint.socket_pool[self.name].send_json(data)

    async def on_disconnect(self, websocket: WebSocket, close_code: int):
        ClientEndpoint.subscriptions[self.name].remove(websocket)  # remove from subscription


@app.websocket(f"{FRONTEND_API_ROOT}" + "/ws/")
async def handle_viewer_websocket(websocket: WebSocket, name: str):
    await ViewerEndpoint(websocket.scope, websocket.receive, websocket.send, name)


class ClientList(BaseModel):
    names: List[str]


@app.get(f"{FRONTEND_API_ROOT}" + "/list", response_model=ClientList)
async def return_names_of_status():
    return client_manager.get_all().keys()


@app.get(f"{FRONTEND_API_ROOT}" + "/status/", response_model=Status)
async def return_status_of(name: str):
    try:
        client = client_manager.get(name)
    except ValueError:
        raise HTTPException(status_code=404, detail="Client not found.")
    return client.status


if __name__ == "__main__":
    cfg = {"port": 5000, "host": ""}
    uvicorn.run("_fastapi_server:app", host=cfg["host"], port=cfg["port"], reload=True)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230319

import inspect
import types
from os import path
from typing import Union


def get_frame_traceback(traceback=1):
    stack = inspect.stack()[traceback]
    return stack


def get_frame_func_name_traceback(traceback=1):
    frame = get_frame_traceback(traceback + 1)
    func = frame.function
    return None if func == "<module>" else func


def get_frame_class_traceback(traceback=1):
    frame = get_frame_traceback(traceback + 1)
    _locals = frame[0].f_locals
    if "self" in _locals:
        return _locals["self"].__class__
    return None


def get_frame_module_traceback(traceback=1) -> Union[types.ModuleType, None]:
    frame = get_frame_traceback(traceback + 1)
    module = inspect.getmodule(frame[0])
    return module


def get_frame_filepath_traceback(traceback=1):
    frame = get_frame_traceback(traceback + 1)
    return frame.filename


class TracedIdentity:
    def __init__(self, frame=None) -> None:
        self.frame = None
        self.func_name = None
        self.locals = None
        self.class_obj = None
        self.class_name = None
        self.module = None
        self.module_name = None
        self.filepath = None
        self.filename = None
        if frame:
            self.parse(frame)

    def parse(self, frame):
        self.frame = frame
        self.func_name = frame.function
        self.locals = frame[0].f_locals
        if "self" in self.locals:
            self.class_obj = self.locals["self"].__class__
            if self.class_obj:
                self.class_name = self.class_obj.__name__
        module: Union[types.ModuleType, None] = inspect.getmodule(frame[0])
        self.module = module
        self.module_name = module.__name__ if module else None
        self.filepath = path.abspath(frame.filename)
        self.filename = path.basename(frame.filename)
        return self

    def __str__(self) -> str:
        return (
            str(self.func_name)
            + ","
            + str(self.class_name)
            + ","
            + str(self.module_name)
            + ","
            + str(self.filepath)
            + ","
        )


def get_caller_identity_traceback(traceback=1):
    frame = get_frame_traceback(traceback + 1)
    traced_identity = TracedIdentity().parse(frame)
    return traced_identity



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230413


import inspect

from neetbox.config._config import DEFAULT_WORKSPACE_CONFIG as default
from neetbox.config._config import get_current
from neetbox.utils.framing import get_frame_module_traceback


def get_module_level_config(module=None):
    try:
        module = (
            module or get_frame_module_traceback(traceback=2).__name__  # type: ignore
        )  # try to trace if module not given
        if type(module) is not str:  # try to trace the belonging module of the given object
            module = inspect.getmodule(module).__name__  # type: ignore
    except Exception:
        module = "@"  # faild to trace, returning all configs
    the_config = get_current()
    sub_module_names = module.split(".")
    if len(sub_module_names):
        if sub_module_names[0] == "neetbox":
            # skip 'neetbox'
            sub_module_names.pop(0)
        for sub_module in sub_module_names:
            if sub_module not in the_config:
                return the_config
            the_config = the_config[sub_module]
    return the_config


__all__ = ["get_module_level_config", "default"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230413


import time
from threading import Thread

import GPUtil
import psutil
from GPUtil import GPU

from neetbox.config import get_module_level_config
from neetbox.integrations import call_on_workspace_load
from neetbox.pipeline import watch
from neetbox.pipeline._signal_and_slot import SYSTEM_CHANNEL
from neetbox.utils import pkg
from neetbox.utils.framing import get_frame_module_traceback
from neetbox.utils.mvc import Singleton

module_name = get_frame_module_traceback().__name__  # type: ignore
assert pkg.is_installed(
    "psutil", try_install_if_not=True
), f"{module_name} requires psutil which is not installed"
assert pkg.is_installed(
    "GPUtil", try_install_if_not=True
), f"{module_name} requires GPUtil which is not installed"


class _CPU_STAT(dict):
    def __init__(self, id=-1, percent=0.0, freq=0.0) -> None:
        self["id"] = id
        self["percent"] = percent
        self["freq"] = freq

    def __str__(self) -> str:
        return f"CPU{self['id']}, {(self['percent'])}%, {(self['freq'])}Mhz"


class _GPU_STAT(dict):
    def __init__(self, id=-1, name="GPU"):
        self["id"] = id
        self["name"] = name

    @classmethod
    def parse(cls, gpu: GPU):
        _instance = _GPU_STAT()
        _instance["id"] = gpu.id
        _instance["name"] = gpu.name
        _instance["load"] = gpu.load
        _instance["memoryUtil"] = gpu.memoryUtil
        _instance["memoryTotal"] = gpu.memoryTotal
        _instance["memoryFree"] = gpu.memoryFree
        _instance["memoryUsed"] = gpu.memoryUsed
        _instance["temperature"] = gpu.temperature
        _instance["driver"] = gpu.driver
        return _instance


class _Hardware(dict, metaclass=Singleton):
    _update_interval = 1.0
    """
    The watcher may interacts with external libraries or devices
    a thread may be more suitable because it can provide a separate execution
    context with its own stack and memory space, which can simplify the design
    and debugging of the code.
    """
    _watcher: Thread = None
    _do_watch: bool = True
    _update_interval: float = 1.0

    def __init__(self) -> None:
        # device
        self["cpus"] = [_CPU_STAT() for _ in range(psutil.cpu_count(logical=True))]
        self["gpus"] = [_GPU_STAT.parse(_gpu) for _gpu in GPUtil.getGPUs()]
        self._with_gpu = False if len(self["gpus"]) == 0 else True
        ram_stat = psutil.virtual_memory()
        self["ram"] = {
            "total": ram_stat[0] / 1e9,
            "available": ram_stat[1] / 1e9,
            "used": ram_stat[3] / 1e9,
            "free": ram_stat[4] / 1e9,
        }
        # the environment shoube be imported in the __init__.py of the outer module. And the watcher thread should be auto started
        # self.set_update_intervel() # do not watch by default

    def json(self):
        return {"cpus": self["cpus"], "ram": self["ram"], "gpus": self["gpus"]}

    def set_update_intervel(self, intervel=1.0) -> None:
        if intervel < 0:
            self._do_watch = False
            return
        self._do_watch = True
        self._update_interval = intervel
        if not self._watcher or not self._watcher.is_alive():

            def watcher_fun(env_instance: _Hardware, do_update_gpus: bool):
                while env_instance._do_watch:
                    # update cpu usage
                    cpu_percent = psutil.cpu_percent(percpu=True)
                    cpu_freq = psutil.cpu_freq(percpu=True)
                    if len(cpu_freq) == 1:
                        cpu_freq = cpu_freq * len(cpu_percent)
                    for index in range(len(cpu_percent)):
                        env_instance["cpus"][index] = _CPU_STAT(
                            id=index,
                            percent=cpu_percent[index],
                            freq=cpu_freq[index],
                        )
                    # update memory usage
                    ram_stat = psutil.virtual_memory()
                    self["ram"] = {
                        "total": ram_stat[0] / 1e9,
                        "available": ram_stat[1] / 1e9,
                        "used": ram_stat[3] / 1e9,
                        "free": ram_stat[4] / 1e9,
                    }
                    # update gpu usage
                    if do_update_gpus:
                        env_instance["gpus"] = [_GPU_STAT.parse(_gpu) for _gpu in GPUtil.getGPUs()]
                    env_instance[""] = psutil.cpu_stats()
                    time.sleep(env_instance._update_interval)

            self._watcher = Thread(target=watcher_fun, args=(self, self._with_gpu), daemon=True)
            self._watcher.start()


# singleton
hardware = _Hardware()


# watch updates in daemon
@call_on_workspace_load(name="hardware-monit")
def load_monit_hardware():
    cfg = get_module_level_config()
    if cfg["monit"]:  # if do monit hardware
        hardware.set_update_intervel(cfg["interval"])

        # watch updates in daemon
        @watch(name="hardware", _channel=SYSTEM_CHANNEL, interval=cfg["interval"])
        def update_env_stat():
            return hardware.json()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# URL:    https://gong.host
# Date:   20230315

import functools
import os
from datetime import date, datetime
from enum import Enum
from random import randint
from typing import Any, Optional, Union

from rich import print as rprint
from rich.panel import Panel

from neetbox.config import get_module_level_config
from neetbox.logging._writer import (
    FileLogWriter,
    JsonLogWriter,
    RawLog,
    consoleLogWriter,
    webSocketLogWriter,
)
from neetbox.logging.formatting import LogStyle, colored_text, styled_text
from neetbox.utils import formatting
from neetbox.utils.framing import get_caller_identity_traceback


class LogLevel(Enum):
    ALL = 4
    DEBUG = 3
    INFO = 2
    WARNING = 1
    ERROR = 0

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value


_GLOBAL_LOG_LEVEL = LogLevel.ALL


def set_log_level(level: LogLevel):
    if type(level) is str:
        level = {
            "ALL": LogLevel.ALL,
            "DEBUG": LogLevel.DEBUG,
            "INFO": LogLevel.INFO,
            "WARNING": LogLevel.WARNING,
            "ERROR": LogLevel.ERROR,
        }[level]
    if type(level) is int:
        assert level >= 0 and level <= 3
        level = LogLevel(level)
    global _GLOBAL_LOG_LEVEL
    _GLOBAL_LOG_LEVEL = level


class Logger:
    # global static
    __WHOM_2_LOGGER = {}
    __WHOM_2_STYLE = {}

    def __init__(self, whom=None, style: Optional[LogStyle] = None):
        self.whom: Any = whom
        if style and style.console_color is None:
            style.randcolor()
        self.style: Optional[LogStyle] = style
        # default writing to console and ws
        self.console_writer = consoleLogWriter
        self.ws_writer = webSocketLogWriter
        self.file_writer = None

        _cfg = get_module_level_config()
        self.set_log_dir(_cfg["logdir"])
        set_log_level(_cfg["level"])

    def __call__(self, whom: Any = None, style: Optional[LogStyle] = None) -> "Logger":
        if whom is None:
            return DEFAULT_LOGGER
        if whom in Logger.__WHOM_2_LOGGER:
            return Logger.__WHOM_2_LOGGER[whom]
        Logger.__WHOM_2_LOGGER[whom] = Logger(whom=whom, style=style)
        return Logger.__WHOM_2_LOGGER[whom]

    def log(
        self,
        *content,
        prefix: Optional[str] = None,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
        traceback=2,
    ):
        _caller_identity = get_caller_identity_traceback(traceback=traceback)

        # converting args into a single string
        _pure_str_message = ""
        for msg in content:
            _pure_str_message += str(msg) + " "

        if type(skip_writers) is str:
            skip_writers = [skip_writers]

        _style = self.style
        if not _style:  # if style not set
            _style_index = str(_caller_identity)
            if _style_index in Logger.__WHOM_2_STYLE:  # check for previous style
                _style = Logger.__WHOM_2_STYLE[_style_index]
            else:
                _style = LogStyle().randcolor()
                Logger.__WHOM_2_STYLE[_style_index] = _style

        raw_log = RawLog(
            rich_msg=_pure_str_message,
            caller_identity=_caller_identity,
            whom=self.whom,
            style=_style,
            prefix=prefix,
            datetime_format=datetime_format,
            with_identifier=with_identifier,
            with_datetime=with_datetime,
            skip_writers=skip_writers,
        )

        for writer in [self.console_writer, self.ws_writer, self.file_writer]:
            if writer:
                raw_log.write_by(writer)

        return self

    def ok(
        self,
        *content,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if _GLOBAL_LOG_LEVEL >= LogLevel.INFO:
            self.log(
                *content,
                prefix=f"[{colored_text('ok', 'green')}]",
                skip_writers=["file", "ws"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
            self.log(
                *content,
                prefix="ok",
                skip_writers=["stdout"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def debug(
        self,
        *content,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if _GLOBAL_LOG_LEVEL >= LogLevel.DEBUG:
            self.log(
                *content,
                prefix=f"[{colored_text('debug', 'cyan')}]",
                skip_writers=["file", "ws"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
            self.log(
                *content,
                prefix="debug",
                skip_writers=["stdout"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def info(
        self,
        *message,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if _GLOBAL_LOG_LEVEL >= LogLevel.INFO:
            self.log(
                *message,
                prefix=f"[{colored_text('info', 'white')}]",
                skip_writers=["file", "ws"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
            self.log(
                *message,
                prefix="info",
                skip_writers=["stdout"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def warn(
        self,
        *message,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
    ):
        if _GLOBAL_LOG_LEVEL >= LogLevel.WARNING:
            self.log(
                *message,
                prefix=f"[{colored_text('warning', 'yellow')}]",
                skip_writers=["file", "ws"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
            self.log(
                *message,
                prefix="warning",
                skip_writers=["stdout"],
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        return self

    def err(
        self,
        err,
        datetime_format: Optional[str] = None,
        with_identifier: Optional[bool] = None,
        with_datetime: Optional[bool] = None,
        skip_writers: list[str] = [],
        reraise=False,
    ):
        if type(err) is not Exception:
            err = RuntimeError(str(err))
        if _GLOBAL_LOG_LEVEL >= LogLevel.ERROR:
            self.log(
                str(err),
                prefix=f"[{colored_text('error','red')}]",
                skip_writers=["file", "ws"] + skip_writers,
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
            self.log(
                str(err),
                prefix="error",
                skip_writers=["stdout"],
                traceback=3,
                datetime_format=datetime_format,
                with_identifier=with_identifier,
                with_datetime=with_datetime,
            )
        if reraise or _GLOBAL_LOG_LEVEL >= LogLevel.DEBUG:
            raise err
        return self

    def mention(self, func):
        @functools.wraps(func)
        def with_logging(*args, **kwargs):
            self.log(f"Currently running: {func.__name__}", traceback=3)
            return func(*args, **kwargs)

        return with_logging

    def console_banner(self, text, font: Optional[str] = None):
        from pyfiglet import Figlet, FigletFont

        builtin_font_list = [
            "ansiregular",
            "ansishadow",
            "isometrixc2",
            "nscripts",
            "nvscript",
        ]
        if not font:
            font = builtin_font_list[randint(0, len(builtin_font_list)) - 1]

        if font not in FigletFont.getFonts():
            if font in builtin_font_list:  # builtin but not installed
                module_path = os.path.dirname(__file__)
                FigletFont.installFonts(f"{module_path}/flfs/{font}.flf")
            else:  # path?
                assert os.path.isfile(
                    font
                ), "The provided font is not a fontname or a font file path."
                file_name = os.path.basename(font)
                file = os.path.splitext(file_name)
                if file[0] not in FigletFont.getFonts():  # no installed file match the file name
                    try:
                        self.info(f"{file[0]} is not installed. Trying to install as a fontfile.")
                        FigletFont.installFonts(f"res/flfs/{font}.flf")
                    except Exception:
                        self.err("Could not install font {font}. Fallback to default.")
                        font = None
                else:
                    font = file[0]
        f = Figlet(font)
        rendered_text = f.renderText(text)
        rprint(Panel.fit(f"{rendered_text}", border_style="green"))
        return self

    def skip_lines(self, line_cnt=1, skip_writers: list[str] = []):
        """Let the logger log some empty lines

        Args:
            line_cnt (int, optional): how many empty line. Defaults to 1.

        Returns:
            _Logger : the logger instance itself
        """
        self.log(
            "\n" * line_cnt, with_datetime=False, with_identifier=False, skip_writers=skip_writers
        )
        return self

    def log_txt_file(self, file, skip_writers: list[str] = []):
        if isinstance(file, str):
            file = open(file)
        context = ""
        for line in file.readlines():
            context += line
        self.log(context, with_datetime=False, with_identifier=False, skip_writers=skip_writers)
        return self

    def set_log_dir(self, path, independent=False):
        if not path:
            self._bind_file(None)
            return self
        if not path:
            self._bind_file(None)
            return self
        if os.path.isfile(path):
            raise Exception("Target path is not a directory.")
        if not os.path.exists(path):
            DEFAULT_LOGGER.info(f"Directory {path} not found, trying to create.")
            try:
                os.makedirs(path)
            except Exception:
                DEFAULT_LOGGER.err(f"Failed when trying to create directory {path}")
                raise Exception(f"Failed when trying to create directory {path}")
        log_file_name = ""
        if independent:
            log_file_name += self.whom
        log_file_name += str(date.today()) + ".log"
        self._bind_file(os.path.join(path, log_file_name))
        return self

    def _bind_file(self, path):
        if not path:
            self.file_writer = None
            return self
        self.file_writer = FileLogWriter(path=path)
        return self

    def file_bend(self) -> bool:
        return self.file_writer is not None


DEFAULT_LOGGER = Logger(None)
